@@ -13,7 +13,7 @@ getLogger("").setLevel(DEBUG)
 
 
 def init(filename=Path("mdpdf.log")):
-    filehandler = FileHandler(filename, mode="w")
+    filehandler = FileHandler(filename, mode="w", encoding="utf-8")
     filehandler.setLevel(DEBUG)
     filehandler.setFormatter(Formatter(LONGFORMAT))
     getLogger("").addHandler(filehandler)
