@@ -1,10 +1,10 @@
-# inspired from https://github.com/readthedocs/commonmark.py/blob/master/commonmark/render/html.py
+# inspired from
+# https://github.com/readthedocs/commonmark.py/blob/master/commonmark/render/html.py
 # https://github.com/readthedocs/commonmark.py/blob/master/commonmark/render/renderer.py
 
 import re
 from builtins import str
 import fitz
-import sys
 from pathlib import Path
 
 
@@ -15,7 +15,6 @@ from . import properties
 from .headfoot import Header, Footer
 
 
-width, height = fitz.PaperSize("letter")  # choose paper format
 fontSize = 10  # choose font size of text
 headingfontSizes = [18, 16, 14, 12, 10, 10, 10, 10]
 lineheight = fontSize * 1.2  # line height is 20% larger
@@ -33,12 +32,16 @@ imageWidthRe = re.compile(r"(.*){.*width\s*=\s*(\d+)%.*}")
 class PdfRenderer:
     def __init__(self, pdf):
         self.list_data = list()  # to store the states of ordered/unordered list
-        self.indir = None  # TODO: put this in render() Directory containing markdown (and images)
+        # TODO: put this in render() Directory containing markdown (and images)
+        self.indir = None
         self.pdf = pdf
         self.doc = fitz.open()
         self.toc = []
         self.currentPage = None
 
+        global width, height
+        width, height = fitz.PaperSize(properties.paperSize)  # choose paper format
+
     def __del__(self):
 
         # Close the file.  If an exception occured, the attribute might not be present,
@@ -51,24 +54,21 @@ class PdfRenderer:
                 try:
                     self.doc.setToC(self.toc)
                 except ValueError as e:
-                    log.info(f"Bad heading level.  More information: ")
+                    log.exception("Bad heading level.  More information:")
                     lastspace = str(e).rfind(" ")
                     log.info(self.toc[int(str(e)[lastspace:])])
                 self.doc.setMetadata(properties.document)
                 try:
                     self.doc.save(str(self.pdf), garbage=4, deflate=True)
                 except RuntimeError as e:
-                    print("e")
+                    log.exception(e)
                 self.doc.close()
             else:
                 log.info("No pages to save")
 
-    def render(self, ast, indir):
-        """Walks the AST and calls member methods for each Node type.
-
-        @param ast {Node} The root of the abstract syntax tree.
-        """
-        self.indir = indir
+    def render(self, ast, inputFile):
+        self.infile = Path(inputFile).absolute()
+        self.indir = Path(inputFile).parent.resolve()
         for node, entering in ast.walker():
             getattr(self, node.t)(node, entering)
 
@@ -107,20 +107,23 @@ class PdfRenderer:
             suffix = line
             while True:
                 index = line.rfind(" ", 0, len(prefix))
-                if index == -1:
+                if index == -1:  # No spaces found.  Print it on the next line
+                    self.cr(chr(0xAC))
+                    self.printSegment(suffix)
                     break
                 prefix = line[:index]
                 prefixWidth = fitz.getTextlength(
                     prefix, fontname=fontName, fontsize=fontSize
                 )
-                if prefixWidth > budget:
+                if prefixWidth > budget:  # Still too large.  Chop again.
                     continue
                 else:
                     self.printSegment(prefix)
                     suffix = line[index + 1 :]
+                    # Recurse for the remainder.  It might be too long as well.
+                    self.cr(chr(0xAC))
+                    self.printLine(suffix)
                     break
-            self.cr(chr(0xAC))
-            self.printSegment(suffix)
 
     # Softbreaks are just CRs in the input, within paragraph
     # it becomes a space.
@@ -162,7 +165,7 @@ class PdfRenderer:
                 self.linkRects.clear()
             else:
                 filename = Path(self.indir) / self.linkDestination
-                if filename.exists():
+                try:
 
                     # This is ugly but whatever
                     pin = fitz.Point(
@@ -171,14 +174,16 @@ class PdfRenderer:
                     self.currentPage.addFileAnnot(
                         pin,
                         buffer=open(filename, mode="rb").read(),
-                        filename=node.first_child.literal,  # This goes in Description  TODO: file issue with PyMuPDF
-                        ufilename=filename.name,  #
-                        desc="deschere",  # Not working.  See TODO: file issue with PyMuPDF
+                        # This goes in Description  TODO: file issue with PyMuPDF
+                        filename=node.first_child.literal,
+                        ufilename=filename.name,
+                        # Not working.  See TODO: file issue with PyMuPDF
+                        desc="deschere",
                         icon="Paperclip",
                     )
 
-                    ## This doesn't work but keeping as comments for future reference.
-                    ## (Linking to an embedded PDF only works on PDF >= 1.6)
+                    # This doesn't work but keeping as comments for future reference.
+                    # (Linking to an embedded PDF only works on PDF >= 1.6)
                     # rc = self.doc.embeddedFileAdd(
                     #     self.linkDestination,  # entry identifier
                     #     buffer=open(filename, mode="rb").read(),
@@ -197,20 +202,30 @@ class PdfRenderer:
                     #     }
                     #     self.currentPage.insertLink(newLink)
 
-                else:
-                    # raise Exception(f"file {filename} not found")
-                    log.info(f"Warning: file {filename} not found")
+                except FileNotFoundError as err:
+                    self.markdownError(node, f"{node.destination}: {err.strerror}")
 
             self.linkDestination = None
 
+    def findSource(self, node):
+        if node.sourcepos:
+            return node.sourcepos[1][0]
+        else:
+            return self.findSource(node.parent)
+
+    def markdownError(self, node, msg):
+        log.error(f"{self.infile}:{self.findSource(node)}: {msg}")
+
     def image(self, node, entering):
         from . import image
 
         if entering:
-            # TODO: consider using the alt attribute to pass the rectangle dimensions/alignment/etc. https://www.w3schools.com/tags/att_img_alt.asp
+            # TODO: consider using the alt attribute to pass the rectangle
+            # dimensions/alignment/etc. https://www.w3schools.com/tags/att_img_alt.asp
             # This means the next node would need to be "peaked" at, since that's how
             # commonmark parses it.
-            # Also, the text callback would have to check if it's the child of an image node, OR, this function would
+            # Also, the text callback would have to check if it's the child of an
+            # image node, OR, this function would
             # jump to the next next child somehow (return??  ) /TODO
 
             try:
@@ -266,11 +281,12 @@ class PdfRenderer:
                         self.insertPoint.y + rectHeight,
                     )
 
-                self.currentPage.insertImage(rect, str(imagefile), keep_proportion=True)
+                self.currentPage.insertImage(
+                    rect, filename=str(imagefile), keep_proportion=True
+                )
                 self.insertPoint.y += rectHeight
             except FileNotFoundError as err:
-                log.info(f"{err}")
-                raise  # TODO: print node/line number
+                self.markdownError(node, f"{node.destination}: {err.strerror}")
 
         # node.title
 
@@ -293,7 +309,8 @@ class PdfRenderer:
                     return
             if node.parent.parent is not None:
                 if node.parent.parent.t == "list":
-                    # if grandparent.list_data["tight"]:  # TODO maybe deal with tight/loose lists
+                    # if grandparent.list_data["tight"]:
+                    # TODO maybe deal with tight/loose lists
                     return
             self.crHalfLine("paragraphhalf+")
         else:
@@ -304,12 +321,18 @@ class PdfRenderer:
 
     def heading(self, node, entering):
         global fontSize
+
+        if not hasattr(node.first_child, "literal"):
+            self.markdownError(node, "Empty heading")
+            return
+
         if entering:
             style.push(fontname=font.HELVETICA, fontsize=headingfontSizes[node.level])
+            lineheight = style.currentStyle().lineheight
             if self.insertPoint.y > margin + lineheight:
                 self.cr("H+")
             if node.level == 1:
-                properties.heading = node.first_child.literal
+                properties.setSection(node.first_child.literal)
         else:
             self.toc.append(
                 [
@@ -321,6 +344,7 @@ class PdfRenderer:
             )
             self.cr("H-")
             style.pop()
+            lineheight = style.currentStyle().lineheight
 
     def code(self, node, entering):
         style.push(fontname=font.COURIER)
@@ -381,15 +405,13 @@ class PdfRenderer:
 
     def item(self, node, entering):
         # attrs = self.attrs(node)
-        # node.sourcepos
         if entering:
             if self.list_data[-1]["type"] == "ordered":
                 self.printSegment(f" {node.list_data['start']} ")
             else:
                 style.push(fontname=font.ZAPFDINGBATS)
-                self.printSegment(
-                    "l"
-                )  # Bullet:  https://help.adobe.com/en_US/framemaker/2015/using/using-framemaker-2015/Appendix/frm_character_sets_cs/frm_character_sets_cs-5.htm
+                self.printSegment("l")  # Bullet:
+                #  https://help.adobe.com/en_US/framemaker/2015/using/using-framemaker-2015/Appendix/frm_character_sets_cs/frm_character_sets_cs-5.htm
                 style.pop()
             self.indent += 16
             self.insertPoint.x = margin + self.indent
@@ -419,6 +441,8 @@ class PdfRenderer:
 
     def cr(self, note):
         # self.currentPage.insertText(self.insertPoint, note, fontsize=3)
+        lineheight = style.currentStyle().lineheight
+
         self.insertPoint.x = margin + self.indent
         self.insertPoint.y += lineheight
         if self.insertPoint.y > height - margin:
@@ -483,12 +507,12 @@ class PdfRenderer:
         style.push(fontname=font.HELVETICA, fontsize=8, indent=0)
         if Header.enabled:
             self.insertPoint.y = margin / 2 + lineheight / 2
-            l = Header.left(properties)
-            m = Header.mid(properties)
-            r = Header.right(properties)
-            self.printLeft(l)
-            self.printCentre(m)
-            self.printRight(r)
+            left = Header.left(properties)
+            middle = Header.mid(properties)
+            right = Header.right(properties)
+            self.printLeft(left)
+            self.printCentre(middle)
+            self.printRight(right)
 
             pntFrom = fitz.Point(margin, 0.75 * margin)
             pntTo = fitz.Point(width - margin, 0.75 * margin)
@@ -496,12 +520,12 @@ class PdfRenderer:
 
         if Footer.enabled:
             self.insertPoint.y = height - (margin / 2) + lineheight / 2
-            l = Footer.left(properties)
-            m = Footer.mid(properties)
-            r = Footer.right(properties)
-            self.printLeft(l)
-            self.printCentre(m)
-            self.printRight(r)
+            left = Footer.left(properties)
+            middle = Footer.mid(properties)
+            right = Footer.right(properties)
+            self.printLeft(left)
+            self.printCentre(middle)
+            self.printRight(right)
 
             pntFrom = fitz.Point(margin, height - (0.75 * margin))
             pntTo = fitz.Point(width - margin, height - (0.75 * margin))
@@ -515,9 +539,6 @@ class PdfRenderer:
 
     def printLeft(self, line):
         sty = style.currentStyle()
-        lineWidth = fitz.getTextlength(
-            line, fontname=sty.font.name, fontsize=sty.fontsize
-        )
         self.insertPoint.x = margin
         self.currentPage.insertText(
             self.insertPoint, line, fontname=sty.font.name, fontsize=sty.fontsize
@@ -542,4 +563,3 @@ class PdfRenderer:
         self.currentPage.insertText(
             self.insertPoint, line, fontname=sty.font.name, fontsize=sty.fontsize
         )
-
